import unittest

from unittest.mock import patch

from data import ImageFolder

import torch

from torchvision.transforms.v2 import Compose


class TestImageFolder(unittest.TestCase):
    @patch("data.walk")
    @patch("data.path.join")
    def setUp(self, mock_path_join, mock_walk):
        mock_walk.return_value = [
            ("/root", ("subdir",), ("image1.png", "image2.jpg", "not_image.txt"))
        ]

        mock_path_join.side_effect = lambda a, b: f"{a}/{b}"

        self.dataset = ImageFolder("/root", 4, 128)

    def test_initialization(self):
        self.assertEqual(len(self.dataset.image_paths), 2)

        self.assertIsInstance(self.dataset.input_transformer, Compose)
        self.assertIsInstance(self.dataset.target_transformer, Compose)

    def test_has_image_extension(self):
        self.assertTrue(self.dataset.has_image_extension("image.png"))
        self.assertFalse(self.dataset.has_image_extension("document.pdf"))

    @patch("data.decode_image")
    def test_getitem(self, mock_decode_image):
        mock_decode_image.return_value = torch.rand((3, 128, 128))

        x, y = self.dataset[0]

        self.assertIsInstance(x, torch.Tensor)
        self.assertIsInstance(y, torch.Tensor)

        self.assertEqual(x.shape, (3, 32, 32))
        self.assertEqual(y.shape, (3, 128, 128))

    def test_len(self):
        self.assertEqual(len(self.dataset), 2)


if __name__ == "__main__":
    unittest.main()



from functools import partial

import torch

from torch import Tensor

from torch.nn import (
    Module,
    ModuleList,
    Conv2d,
    Sigmoid,
    SiLU,
    Upsample,
    PixelShuffle,
)

from torch.nn.utils.parametrizations import weight_norm
from torch.nn.utils.parametrize import remove_parametrizations
from torch.utils.checkpoint import checkpoint as torch_checkpoint

from huggingface_hub import PyTorchModelHubMixin


class UltraZoom(Module, PyTorchModelHubMixin):
    """
    A fast single-image super-resolution model with a deep low-resolution encoder network
    and high-resolution sub-pixel convolutional decoder head with global residual pathway.

    Ultra Zoom uses a "zoom in and enhance" approach to upscale images by first increasing
    the resolution of the input image using bicubic interpolation and then filling in the
    details using a deep neural network.
    """

    AVAILABLE_UPSCALE_RATIOS = {2, 4, 8}

    AVAILABLE_HIDDEN_RATIOS = {1, 2, 4}

    def __init__(
        self,
        upscale_ratio: int,
        num_channels: int,
        hidden_ratio: int,
        num_encoder_layers: int,
    ):
        super().__init__()

        if upscale_ratio not in self.AVAILABLE_UPSCALE_RATIOS:
            raise ValueError(
                f"Upscale ratio must be either 2, 4, or 8, {upscale_ratio} given."
            )

        self.skip = Upsample(scale_factor=upscale_ratio, mode="bicubic")

        self.encoder = Encoder(num_channels, hidden_ratio, num_encoder_layers)

        self.decoder = SubpixelConv2d(num_channels, upscale_ratio)

        self.upscale_ratio = upscale_ratio

    @property
    def num_params(self) -> int:
        """Total number of parameters in the model."""
        
        return sum(param.numel() for param in self.parameters())

    @property
    def num_trainable_params(self) -> int:
        return sum(param.numel() for param in self.parameters() if param.requires_grad)

    def add_weight_norms(self) -> None:
        """Add weight normalization to all Conv2d layers in the model."""

        for module in self.modules():
            if isinstance(module, Conv2d):
                weight_norm(module)

    def remove_weight_norms(self) -> None:
        """Remove weight normalization parameterization."""

        for module in self.modules():
            if isinstance(module, Conv2d) and hasattr(module, "parametrizations"):
                params = [name for name in module.parametrizations.keys()]

                for name in params:
                    remove_parametrizations(module, name)

    def forward(self, x: Tensor) -> tuple[Tensor, Tensor]:
        s = self.skip.forward(x)

        z = self.encoder.forward(x)
        z = self.decoder.forward(z)

        z = s + z  # Global residual connection

        return z, s

    @torch.no_grad()
    def test_compare(self, x: Tensor) -> Tensor:
        z, s = self.forward(x)

        z = torch.clamp(z, 0, 1)
        s = torch.clamp(s, 0, 1)

        return z, s

    @torch.no_grad()
    def upscale(self, x: Tensor) -> Tensor:
        z, _ = self.forward(x)

        z = torch.clamp(z, 0, 1)

        z *= 255

        z = z.to(torch.uint8)

        return z


class Encoder(Module):
    """A low-resolution subnetwork employing a deep stack of encoder blocks."""

    def __init__(self, num_channels: int, hidden_ratio: int, num_layers: int):
        super().__init__()

        assert num_layers > 0, "Number of layers must be greater than 0."

        self.stem = Conv2d(3, num_channels, kernel_size=3, padding=1)

        self.body = ModuleList(
            [EncoderBlock(num_channels, hidden_ratio) for _ in range(num_layers)]
        )

        self.checkpoint = lambda layer, x: layer(x)

    def enable_activation_checkpointing(self) -> None:
        """
        Instead of memorizing the activations of the forward pass, recompute them
        at every encoder block.
        """

        self.checkpoint = partial(torch_checkpoint, use_reentrant=False)

    def forward(self, x: Tensor) -> Tensor:
        z = self.stem.forward(x)

        for layer in self.body:
            z = self.checkpoint(layer, z)

        return z


class EncoderBlock(Module):
    """A single encoder block consisting of two stages and a residual connection."""

    def __init__(self, num_channels: int, hidden_ratio: int):
        super().__init__()

        self.stage1 = SpatialAttention(num_channels)
        self.stage2 = InvertedBottleneck(num_channels, hidden_ratio)

    def forward(self, x: Tensor) -> Tensor:
        z = self.stage1.forward(x)
        z = self.stage2.forward(z)

        z = x + z  # Local residual connection

        return z


class SpatialAttention(Module):
    """A spatial attention module with large depth-wise convolutions."""

    def __init__(self, num_channels: int):
        super().__init__()

        assert num_channels > 0, "Number of channels must be greater than 0."

        self.depthwise = Conv2d(
            num_channels,
            num_channels,
            kernel_size=11,
            padding=5,
            groups=num_channels,
            bias=False,
        )

        self.pointwise = Conv2d(num_channels, num_channels, kernel_size=1)

        self.sigmoid = Sigmoid()

    def forward(self, x: Tensor) -> Tensor:
        z = self.depthwise.forward(x)
        z = self.pointwise.forward(z)

        z = self.sigmoid.forward(z)

        z = z * x

        return z


class InvertedBottleneck(Module):
    """A wide non-linear activation block with 3x3 convolutions."""

    def __init__(self, num_channels: int, hidden_ratio: int):
        super().__init__()

        assert num_channels > 0, "Number of channels must be greater than 0."
        assert hidden_ratio in {1, 2, 4}, "Hidden ratio must be either 1, 2, or 4."

        hidden_channels = hidden_ratio * num_channels

        self.conv1 = Conv2d(num_channels, hidden_channels, kernel_size=3, padding=1)
        self.conv2 = Conv2d(hidden_channels, num_channels, kernel_size=3, padding=1)

        self.silu = SiLU()

    def forward(self, x: Tensor) -> Tensor:
        z = self.conv1.forward(x)
        z = self.silu.forward(z)
        z = self.conv2.forward(z)

        return z


class SubpixelConv2d(Module):
    """A high-resolution decoder using sub-pixel convolution."""

    def __init__(self, in_channels: int, upscale_ratio: int):
        super().__init__()

        assert upscale_ratio in {2, 4, 8}, "Upscale ratio must be either 2, 4, or 8."

        out_channels = 3 * upscale_ratio**2

        self.conv = Conv2d(in_channels, out_channels, kernel_size=3, padding=1)

        self.shuffle = PixelShuffle(upscale_ratio)

    def forward(self, x: Tensor) -> Tensor:
        z = self.conv.forward(x)
        z = self.shuffle.forward(z)

        return z



from unittest import main, TestCase

import torch

from torch import Tensor

from loss import TVLoss


class TestTVLoss(TestCase):
    def setUp(self):
        self.loss_function = TVLoss()

    def test_tv_loss_initialization(self):
        self.assertIsInstance(self.loss_function, TVLoss)

    def test_tv_loss_forward(self):
        y_pred = torch.rand((1, 3, 4, 4))

        penalty = self.loss_function(y_pred)

        self.assertIsInstance(penalty, Tensor)
        self.assertGreaterEqual(penalty.item(), 0)

    def test_tv_loss_forward_zero(self):
        y_pred = torch.zeros((1, 3, 4, 4))

        penalty = self.loss_function(y_pred)

        self.assertEqual(penalty.item(), 0)


if __name__ == "__main__":
    main()
